# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: renderserver.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="renderserver.proto",
    package="renderserver",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x12renderserver.proto\x12\x0crenderserver"f\n\x16UpdateSceneDataRequest\x12"\n\x05scene\x18\x01 \x01(\x0b\x32\x13.renderserver.Scene\x12\x11\n\texception\x18\x02 \x01(\t\x12\x15\n\rhas_exception\x18\x03 \x01(\x08"\\\n\x05Scene\x12\x0c\n\x04name\x18\x01 \x01(\t\x12+\n\nanimations\x18\x02 \x03(\x0b\x32\x17.renderserver.Animation\x12\x18\n\x10\x62\x61\x63kground_color\x18\x03 \x01(\t"+\n\tAnimation\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64uration\x18\x02 \x01(\x02"\x0e\n\x0c\x45mptyRequest"\x0f\n\rEmptyResponse2d\n\x0cRenderServer\x12T\n\x0fUpdateSceneData\x12$.renderserver.UpdateSceneDataRequest\x1a\x1b.renderserver.EmptyResponseb\x06proto3',
)


_UPDATESCENEDATAREQUEST = _descriptor.Descriptor(
    name="UpdateSceneDataRequest",
    full_name="renderserver.UpdateSceneDataRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="scene",
            full_name="renderserver.UpdateSceneDataRequest.scene",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="exception",
            full_name="renderserver.UpdateSceneDataRequest.exception",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="has_exception",
            full_name="renderserver.UpdateSceneDataRequest.has_exception",
            index=2,
            number=3,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=36,
    serialized_end=138,
)


_SCENE = _descriptor.Descriptor(
    name="Scene",
    full_name="renderserver.Scene",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="renderserver.Scene.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="animations",
            full_name="renderserver.Scene.animations",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="background_color",
            full_name="renderserver.Scene.background_color",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=140,
    serialized_end=232,
)


_ANIMATION = _descriptor.Descriptor(
    name="Animation",
    full_name="renderserver.Animation",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="renderserver.Animation.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="duration",
            full_name="renderserver.Animation.duration",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=234,
    serialized_end=277,
)


_EMPTYREQUEST = _descriptor.Descriptor(
    name="EmptyRequest",
    full_name="renderserver.EmptyRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=279,
    serialized_end=293,
)


_EMPTYRESPONSE = _descriptor.Descriptor(
    name="EmptyResponse",
    full_name="renderserver.EmptyResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=295,
    serialized_end=310,
)

_UPDATESCENEDATAREQUEST.fields_by_name["scene"].message_type = _SCENE
_SCENE.fields_by_name["animations"].message_type = _ANIMATION
DESCRIPTOR.message_types_by_name["UpdateSceneDataRequest"] = _UPDATESCENEDATAREQUEST
DESCRIPTOR.message_types_by_name["Scene"] = _SCENE
DESCRIPTOR.message_types_by_name["Animation"] = _ANIMATION
DESCRIPTOR.message_types_by_name["EmptyRequest"] = _EMPTYREQUEST
DESCRIPTOR.message_types_by_name["EmptyResponse"] = _EMPTYRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UpdateSceneDataRequest = _reflection.GeneratedProtocolMessageType(
    "UpdateSceneDataRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATESCENEDATAREQUEST,
        "__module__": "renderserver_pb2"
        # @@protoc_insertion_point(class_scope:renderserver.UpdateSceneDataRequest)
    },
)
_sym_db.RegisterMessage(UpdateSceneDataRequest)

Scene = _reflection.GeneratedProtocolMessageType(
    "Scene",
    (_message.Message,),
    {
        "DESCRIPTOR": _SCENE,
        "__module__": "renderserver_pb2"
        # @@protoc_insertion_point(class_scope:renderserver.Scene)
    },
)
_sym_db.RegisterMessage(Scene)

Animation = _reflection.GeneratedProtocolMessageType(
    "Animation",
    (_message.Message,),
    {
        "DESCRIPTOR": _ANIMATION,
        "__module__": "renderserver_pb2"
        # @@protoc_insertion_point(class_scope:renderserver.Animation)
    },
)
_sym_db.RegisterMessage(Animation)

EmptyRequest = _reflection.GeneratedProtocolMessageType(
    "EmptyRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTYREQUEST,
        "__module__": "renderserver_pb2"
        # @@protoc_insertion_point(class_scope:renderserver.EmptyRequest)
    },
)
_sym_db.RegisterMessage(EmptyRequest)

EmptyResponse = _reflection.GeneratedProtocolMessageType(
    "EmptyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTYRESPONSE,
        "__module__": "renderserver_pb2"
        # @@protoc_insertion_point(class_scope:renderserver.EmptyResponse)
    },
)
_sym_db.RegisterMessage(EmptyResponse)


_RENDERSERVER = _descriptor.ServiceDescriptor(
    name="RenderServer",
    full_name="renderserver.RenderServer",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=312,
    serialized_end=412,
    methods=[
        _descriptor.MethodDescriptor(
            name="UpdateSceneData",
            full_name="renderserver.RenderServer.UpdateSceneData",
            index=0,
            containing_service=None,
            input_type=_UPDATESCENEDATAREQUEST,
            output_type=_EMPTYRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_RENDERSERVER)

DESCRIPTOR.services_by_name["RenderServer"] = _RENDERSERVER

# @@protoc_insertion_point(module_scope)
